package writecache

import (
	"context"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	storagelog "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/internal/log"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/internal/metaerr"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/kvio"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
)

// Delete removes object from write-cache.
//
// Returns an error of type apistatus.ObjectNotFound if object is missing in write-cache.
func (c *cache) Delete(ctx context.Context, addr oid.Address) error {
	ctx, span := tracing.StartSpanFromContext(ctx, "writecache.Delete",
		trace.WithAttributes(
			attribute.String("address", addr.EncodeToString()),
		))
	defer span.End()

	deleted := false
	storageType := StorageTypeUndefined
	startedAt := time.Now()
	defer func() {
		c.metrics.Delete(time.Since(startedAt), deleted, storageType)
	}()

	c.modeMtx.RLock()
	defer c.modeMtx.RUnlock()
	if c.readOnly() {
		return ErrReadOnly
	}

	saddr := addr.EncodeToString()

	var dataSize int
	_ = c.repo.Read(func(tx kvio.ReadOnlyTx) error {
		return tx.Read([]byte(saddr), func(v kvio.Value) error {
			dataSize = len(v)
			return nil
		})
	})

	if dataSize > 0 {
		storageType = StorageTypeDB
		if err := kvio.Delete(c.repo, []byte(saddr)); err != nil {
			return err
		}
		storagelog.Write(c.log,
			storagelog.AddressField(saddr),
			storagelog.StorageTypeField(wcStorageType),
			storagelog.OpField("db DELETE"),
		)
		deleted = true
		c.objCounters.DecDB()
		return nil
	}

	storageType = StorageTypeFSTree
	_, err := c.fsTree.Delete(ctx, common.DeletePrm{Address: addr})
	if err == nil {
		storagelog.Write(c.log,
			storagelog.AddressField(saddr),
			storagelog.StorageTypeField(wcStorageType),
			storagelog.OpField("fstree DELETE"),
		)
		c.objCounters.DecFS()
		deleted = true
	}

	return metaerr.Wrap(err)
}
