package writecache

import (
	"context"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/internal/metaerr"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/kvio"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	objectSDK "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
)

// Get returns object from write-cache.
//
// Returns an error of type apistatus.ObjectNotFound if the requested object is missing in write-cache.
func (c *cache) Get(ctx context.Context, addr oid.Address) (*objectSDK.Object, error) {
	saddr := addr.EncodeToString()

	ctx, span := tracing.StartSpanFromContext(ctx, "writecache.Get",
		trace.WithAttributes(
			attribute.String("address", saddr),
		))
	defer span.End()

	obj, err := c.getInternal(ctx, saddr, addr)
	return obj, metaerr.Wrap(err)
}

func (c *cache) getInternal(ctx context.Context, saddr string, addr oid.Address) (*objectSDK.Object, error) {
	found := false
	storageType := StorageTypeUndefined
	startedAt := time.Now()
	defer func() {
		c.metrics.Get(time.Since(startedAt), found, storageType)
	}()

	value, err := Get(c.repo, []byte(saddr))
	if err == nil {
		obj := objectSDK.New()
		found = true
		storageType = StorageTypeDB
		return obj, obj.Unmarshal(value)
	}

	res, err := c.fsTree.Get(ctx, common.GetPrm{Address: addr})
	if err != nil {
		return nil, logicerr.Wrap(apistatus.ObjectNotFound{})
	}

	found = true
	storageType = StorageTypeFSTree
	return res.Object, nil
}

// Head returns object header from write-cache.
//
// Returns an error of type apistatus.ObjectNotFound if the requested object is missing in write-cache.
func (c *cache) Head(ctx context.Context, addr oid.Address) (*objectSDK.Object, error) {
	saddr := addr.EncodeToString()

	ctx, span := tracing.StartSpanFromContext(ctx, "writecache.Head",
		trace.WithAttributes(
			attribute.String("address", saddr),
		))
	defer span.End()

	obj, err := c.getInternal(ctx, saddr, addr)
	if err != nil {
		return nil, metaerr.Wrap(err)
	}

	return obj.CutPayload(), nil
}

// Get fetches object from the underlying database.
// Key should be a stringified address.
//
// Returns an error of type apistatus.ObjectNotFound if the requested object is missing in db.
func Get(repo kvio.Repository, key []byte) ([]byte, error) {
	value, err := kvio.Read(repo, key)
	if err != nil {
		return nil, metaerr.Wrap(err)
	}
	if value == nil {
		return nil, metaerr.Wrap(logicerr.Wrap(apistatus.ObjectNotFound{}))
	}
	return value, nil
}
